{% extends "base.html" %}

<!-- Extrahead -->
{% block extrahead %}

  <meta property="og:url" content="{{ page.canonical_url }}">
  {% if page and page.meta and page.meta.title %}
    <meta property="og:title" content="{{ page.meta.title }}">
  {% elif page and page.title and not page.is_homepage %}
    <meta property="og:title" content="{{ page.title }} - {{ config.site_name }}">
  {% else %}
    <meta property="og:title" content="{{ config.site_name }}">
  {% endif %}
  <meta property="og:description" content="{{ config.site_description }}">
  <meta property="og:image" content="https://microsoft.github.io/reconner/img/recon.svg">
  <meta property="og:image:alt" content="Recon NER">

  <meta name="twitter:card" content="summary_large_image">
  <meta name="twitter:site" content="@kabir_khan14">
  <meta name="twitter:creator" content="@kabir_khan14">
  {% if page and page.meta and page.meta.title %}
    <meta property="twitter:title" content="{{ page.meta.title }}">
  {% elif page and page.title and not page.is_homepage %}
    <meta property="twitter:title" content="{{ page.title }} - {{ config.site_name }}">
  {% else %}
    <meta property="twitter:title" content="{{ config.site_name }}">
  {% endif %}
  <meta name="twitter:description" content="{{ config.site_description }}">
  <meta name="twitter:image" content="https://microsoft.github.io/reconner/img/recon.svg">
  <meta name="twitter:image:alt" content="Recon NER">

{% endblock %}